;       AX.25 rx routine                        N0QBH   2000
;       10 Mhz XTAL
;       Ramtron serial 2k RAM installed
;       VISUAL 1 on after 2 flags rx'd
;       Packet storage up to 255 bytes
;       Serial sign-on message
;       Rotating RAMAddressHigh
;       The FRAM and CRC routines are borrrowed from
;       Byon Garrabrant N6BG
;
;       Version: 1.2
;       Date: Mar 30, 2000
;
;------------------------------------------------------------
	LIST	P=PIC16F84
	include	"P16F84.INC"
        __CONFIG _HS_OSC & _CP_OFF & _PWRTE_ON & _WDT_OFF
;-------------------------------------------------------------
;               Port Set-up
;           0   1   2   3   4   5   6   7       bit
; port  A   Out Out Out Out In  X   X   X
;       B   Out Out Out In  In  Out Out Out
;------------------------------------------------------------
	ORG	0			; start code
	goto	main
;------------------------------------------------------------
intro_message
        addwf   PCL,F
        dt      " TURBO PIC-E packet decoder *"
        dt      " N0QBH 2000"
        dt      0x0d,0x0a,0x00
;------------------------------------------------------------
        cblock  0x0c
                aux_count               ;flag counter
                byte_count              ;delay counter
                store_count             ;bytes stored cnt
                serial_count            ;serial out counter
                bit_stuff_count         ;counter
                bit_count               ;counter
                send_reg                ;serial data
                FLAGS                   ;various flags
                delay_count             ;delay timer
                byte_in                 ;data byte rx'd
                AX25_crc_hi             ;upper byte CRC word
                AX25_crc_lo             ;lower byte CRC word
                next_to_last_byte       ;2nd from last byte
                last_byte               ;last byte rxed
                ssid_offset             ;SSID store loc
                ssid_buffer             ;SSID storage
                RAMAddressHigh          ;Upper 3 bits
                RAMAddressLow           ;lower 8 bits
                RAMSave                 ;buffer
                RAMData                 ;data byte
                RAMCount                ;ram counter
        endc
;------------------------------------------------------------
#define RAM_DATA_IN             PORTA,0         ;RAM SO
#define RAM_DATA_OUT            PORTA,0         ;RAM SI
#define SERIAL_DATA_OUT         PORTA,1         ;4800 out
#define OPT_TOGGLE              PORTA,4         ;SW 1 in
#define VISUAL_ONE              PORTB,0         ;LED
#define MODE1_OUT               PORTB,1         ;RX/TX mode
#define MODE0_OUT               PORTB,2         ;RX/TX/lo pwr
#define RAM_CLOCK               PORTB,6         ;RAM SCK
#define RAM_CHIP_SELECT         PORTB,7         ;RAM CS
#define DATA_FLAG               FLAGS,1         ;beg/end flag
#define SHIFTED_DATA            FLAGS,2         ;end shft data
#define RAM_WRITE_FLAG          FLAGS,3         ;RAM setup?
;------------------------------------------------------------
#define RAM_OP_WREN             B'00000110'
#define	RAM_OP_READ		B'00000011'
#define	RAM_OP_WRITE  		B'00000010'
;------------------------------------------------------------
main    
        bsf     STATUS,RP0              ;bank 1
        movlw   b'11110000'             ;A:4 input
        movwf   TRISA                   ;set up port A I/O
        movlw   b'00011000'             ;B:3,:4,inputs
        movwf   TRISB                   ;set up port B I/O
        movlw   b'11010011'
        movwf   OPTION_REG              ;TMR0 /16
        bcf     STATUS,RP0              ;bank 0
        clrf    INTCON                  ;clear INT reg
        bsf     MODE0_OUT               ;MX614 setup
        bsf     MODE1_OUT               ;lo pwr mode
        bcf     SERIAL_DATA_OUT         ;stop bit set
        movlw   0x0a                    ;10 loops
        call    adjustable_delay        ;2 mS delay
        bcf     MODE0_OUT               ;MX614 RX mode
        movlw   0x64                    ;100 loops
        call    adjustable_delay        ;20 mS delay
        call    startup_message         ;sign on
;------------------------------------------------------------
house_clean                             ;new packet
        call    RAM_init                ;init RAM
        clrf    aux_count               ;reset misc counter
        clrf    byte_in                 ;clear data byte
        clrf    bit_stuff_count         ;reset b_s_count
        clrf    store_count             ;reset count out
        clrf    byte_count              ;reset byte count
        clrf    FLAGS                   ;reset flags
        movlw   0xff
        movwf   AX25_crc_hi             ;reset CRC calc
        movwf   AX25_crc_lo
        movlw   0x7e
        movwf   TMR0
        movfw   PORTB                   ;reset bit change
        clrf    INTCON                  ;set up TMR0, RBI
        bcf     VISUAL_ONE              ;clear flags lite
find_start_flag
        clrc
        call    recv_one_bit            ;check for toggle
        skpnz
        setc                            ;no toggle (1)
        rrf     byte_in,F               ;toggle  (0)
        movfw   byte_in                 ;byte_in to W
        xorlw   0x7e                    ;is it a flag ?
        skpz
        goto    find_start_flag         ;no
start_flag_loop                         ;yes, look again
        call    recv_one_bit
        skpnz                           ;see a toggle ?
        goto    find_start_flag         ;no, go back
        movlw   0x06                    ;yes, must see 6 1's
        movwf   store_count
long_ones
        call    recv_one_bit            ;look for 6
        skpz                            ;1's in arow
        goto    find_start_flag         ;saw a 0
        decfsz  store_count,F           ;saw a 1
        goto    long_ones
        call    recv_one_bit            ;look for end 0
        skpnz
        goto    find_start_flag         ;not flag, reset
        bsf     VISUAL_ONE              ;flag, LED on
;------------------------------------------------------------
;
get_next_byte
        clrf    byte_in                 ;clear input byte
        movlw   0x08
        movwf   bit_count               ;get 8 bits
get_next_bit
        call    recv_one_bit
        skpz
        goto    zero_in                 ; 0 rx'd
        incf    bit_stuff_count,F       ;inc counter
        setc                            ; 1 rx'd
roll_em_in
        rrf     byte_in,F               ;bit stored
        movfw   bit_stuff_count
        xorlw   0x05                    ;5 1's in a row ?
        skpnz
        goto    check_stuff             ;yes
count_em_out                            ;no
        decfsz  bit_count,F             ;8 bits ?
        goto    get_next_bit            ;no
        goto    CRC_save_store          ;yes
zero_in
        clrf    bit_stuff_count         ;0 seen, clear bsc
        clrc
        goto    roll_em_in
;------------------------------------------------------------
; Bit stuff decode routine
;
check_stuff
        call    recv_one_bit            ;wait 1 bit
        skpnz                           ;if data=0, skip next
        goto    sixth_one               ;unchanged, flag ?
        clrf    bit_stuff_count         ;tog seen, clear count
        goto    count_em_out            ;don't count this bit
sixth_one
        call    recv_one_bit            ;get 6th bit
        skpnz                           ;flag ?
        goto    house_clean             ;no, 7th 1, reset
        movfw   bit_count
        xorlw   0x03
        skpz                            ;in the right place ?
        goto    house_clean             ;no
        clrf    bit_stuff_count         ;yes, flag, clear bstc
        btfss   DATA_FLAG               ;header/tail flag ?
        goto    get_next_byte           ;header flag
        movfw   byte_count              ;tail flag
        sublw   0x14                    ;how long is packet ?
        skpnc
        goto    house_clean             ; < 21 bytes rx'd
        goto    packet_end              ;it's long enough
;
CRC_save_store
        btfss   DATA_FLAG               ;test data flag
        bsf     DATA_FLAG               ;set data flag
        incf    byte_count,F            ;count bytes
        movfw   byte_count
        xorlw   0xff                    ;how long ?
        skpnz
        goto    house_clean             ; 255 bytes, reset
        movfw   byte_count              ;how long again
        sublw   0x02
        skpc                            ; >2 ?
        call    AX25_CRC_calc           ;
        movfw   last_byte               ;last byte to W
        movwf   next_to_last_byte       ;to 2nd to last
        movfw   byte_in                 ;get new data byte
        movwf   last_byte               ;store in last byte
        goto    get_next_byte           ;get next byte
;
;------------------------------------------------------------
;       Invert CRC data and compare to value sent
;
packet_end
        movlw   0xff                    ;255 to W
        call    RAMWrite                ;store ff (end)
        call    RAMStop                 ;stop Ram write
        movlw   0xff
        xorwf   AX25_crc_lo,F
        xorwf   AX25_crc_hi,F           ;invert  CRC data
        movfw   next_to_last_byte       ;check lobyte match
        xorwf   AX25_crc_lo,W
        skpz
        goto    house_clean             ;mismatch, reset
        movfw   last_byte
        xorwf   AX25_crc_hi,W           ;check hibyte match
        skpz
        goto    house_clean             ;mismatch, reset
;------------------------------------------------------------
read_back_data                          ;send data serially
        movlw   0x07                    ;org call loc
        movwf   RAMAddressLow           ;in memory
        movlw   0x06
        movwf   aux_count               ;6 loops
        call    RAMReadStart            ;start read at loc
        call    parse_call
        call    RAMStop
        movlw   0x0d                    ;ssid loc
        movwf   ssid_offset
        call    parse_ssid              ;get/send SSID
;
        movlw   '>'                     ;insert r_arrow
        movwf   send_reg
        call    serial_out              ;send >
;
        clrf    RAMAddressLow           ;start of dest call
        movlw   0x06
        movwf   aux_count               ;6 loops
        call    RAMReadStart            ;start read at loc
        call    parse_call
        call    RAMStop
        movlw   0x06
        movwf   ssid_offset
        call    parse_ssid
;
        btfss   SHIFTED_DATA            ;shifted data
        goto    still_shifted           ;yes
        movlw   0x0d                    ;no, do info
        movwf   RAMAddressLow           ;skip over org
        goto    info_field_begin         ;yes
still_shifted
        movlw   ','                     ;no
        movwf   send_reg
        call    serial_out              ;send comma
        movlw   0x0e
        movwf   RAMAddressLow           ;start of next call
        movlw   0x06
        movwf   aux_count               ;6 bytes worth
        movlw   0x14
        movwf   ssid_offset             ;next call ssid loc
display_digipeaters
        call    RAMReadStart            ;start read at loc
        call    parse_call              ;parse / send next
        call    RAMStop
        call    parse_ssid
        btfsc   SHIFTED_DATA
        goto    info_field_begin        ;S_D flag seen
        movlw   ','                     ;no
        movwf   send_reg
        call    serial_out              ;send comma
        movlw   0x07                    ;another address
        addwf   ssid_offset,F           ;next SSID loc
        incf    RAMAddressLow,F         ;next call loc
        movlw   0x06                    ;num of chars
        movwf   aux_count
        goto    display_digipeaters
;
info_field_begin
        movlw   ':'
        movwf   send_reg
        call    serial_out              ;send :
;
        movlw   0x03
        addwf   RAMAddressLow,F         ;loc to read routine
        call    RAMReadStart            ;start read at loc
info_field_loop
        call    RAMRead
        movwf   send_reg                ;W to send_reg
        xorlw   0xff
        skpnz
        goto    wrap_it_up              ;ff, end of data
        call    serial_out              ;send byte
        goto    info_field_loop         ;end not seen
wrap_it_up
        call    RAMStop                 ;end of read
        movlw   0x0d                    ;>, send end
        movwf   send_reg                ;<CR>
        call    serial_out
        movlw   0x0a                    ;<LF>
        movwf   send_reg
        call    serial_out
        goto    house_clean
;------------------------------------------------------------
;       Subroutines area
;------------------------------------------------------------
startup_message
        clrf    aux_count               ;offset clear
startup_loop
        movfw   aux_count               ;offset to W
        call    intro_message           ;get offset byte
        movwf   send_reg                ;store
        xorlw   0x00                    ;test for end
        skpnz
        return                          ;done, start prgm
        call    serial_out              ;send byte
        incf    aux_count,F             ;inc offset
        goto    startup_loop            ;loop
;------------------------------------------------------------
RAM_init
        incf    RAMAddressHigh,F        ;inc high addr byte
        movfw   RAMAddressHigh
        sublw   0x07
        skpc                            ; > 7 ?
        clrf    RAMAddressHigh          ;yes, clear
        clrf    RAMAddressLow           ;no
        clrf    RAMData
        bsf     RAM_CHIP_SELECT
        bcf     RAM_CLOCK
        return
;------------------------------------------------------------
RAMWrite
        bsf     STATUS,RP0              ;bank 1
        bcf     RAM_DATA_OUT            ;make A:0 output
        bcf     STATUS,RP0              ;bank 0
        movwf   RAMData                 ;W is data
        movlw   D'8'
        movwf   RAMCount                ;8 bits
RAMSendLoop
	rlf	RAMData, F
	bcf	RAM_DATA_OUT
	skpnc
        bsf     RAM_DATA_OUT
	bsf	RAM_CLOCK
	bcf	RAM_CLOCK
	decfsz	RAMCount, F
        goto    RAMSendLoop
	return
;------------------------------------------------------------
RAMRead
        bsf     STATUS,RP0              ;bank 1
        bsf     RAM_DATA_IN             ;make A:0 input
        bcf     STATUS,RP0              ;bank 0
	movlw	D'8'
	movwf	RAMCount
RAMGetLoop
	bsf	RAM_CLOCK
	clrc
	btfsc	RAM_DATA_IN
        setc
	rlf	RAMData, F
	bcf	RAM_CLOCK
	decfsz	RAMCount, F
        goto    RAMGetLoop
	movfw	RAMData
        decfsz  byte_count,F            ;overrun protect
	return
        movlw   0xff                    ;insert end flag
        return
;------------------------------------------------------------
RAMWriteEnable
	bcf	RAM_CHIP_SELECT
	movlw	RAM_OP_WREN
	call	RAMWrite
RAMStop
	bsf	RAM_CHIP_SELECT
	return
;------------------------------------------------------------
RAMWriteStart
	call	RAMWriteEnable
	bcf	RAM_CHIP_SELECT
	swapf	RAMAddressHigh, W
	movwf	RAMData
	rrf	RAMData, W
	andlw	B'00111000'
	iorlw	RAM_OP_WRITE
	call	RAMWrite
	movfw	RAMAddressLow
	call	RAMWrite
	return
;
RAMWriteByte
	movwf	RAMSave
	call	RAMWriteStart
	movfw	RAMSave
	call	RAMWrite
	goto	RAMStop
;------------------------------------------------------------
RAMReadStart
	bcf	RAM_CHIP_SELECT
	swapf	RAMAddressHigh, W
	movwf	RAMData
	rrf	RAMData, W
	andlw	B'00111000'
	iorlw	RAM_OP_READ
	call	RAMWrite
	movfw	RAMAddressLow
	call	RAMWrite
	return
;
RAMReadByte
	call	RAMReadStart
	call	RAMRead
        goto    RAMStop
;------------------------------------------------------------
;       Stores all packet data as received
;
AX25_CRC_calc
        btfsc   RAM_WRITE_FLAG          ;are we setup ?
        goto    FRAM_store              ;yes
        clrf    RAMAddressLow           ;no, start
        call    RAMWriteStart           ;loc to RAM
        bsf     RAM_WRITE_FLAG          ;setup done
;       RAM data store routine
;
FRAM_store
        movfw   next_to_last_byte       ;get byte
        call    RAMWrite                ;store in RAM
;       CRC calculation
crc_rotor
        movlw   0x08
        movwf   bit_count               ; 8 loops
crc_head
        rrf     next_to_last_byte,F     ; get bit for CRC
        call    crc_core                ;do CRC calc
        decfsz  bit_count,F             ;8 counts
        goto    crc_head
        return
;------------------------------------------------------------
crc_core
        movlw   1
        skpnc
        xorwf   AX25_crc_lo, F          ; crc calculation
        clrc
        rrf     AX25_crc_hi, F
        rrf     AX25_crc_lo, F
        skpc
        return
        movlw   H'08'
        xorwf   AX25_crc_lo, F
        movlw   H'84'
        xorwf   AX25_crc_hi, F
        return
;------------------------------------------------------------
;       Assumes RAMReadStart w/address has been called
;
parse_call
        call    RAMRead                 ;get byte from RAM
        movwf   send_reg                ;byte to send_reg
        clrc
        rrf     send_reg,F              ;right shift to normal
        movfw   send_reg
        xorlw   ' '                     ;space ?
        skpnz
        return                          ;yes
        call    serial_out              ;no, send byte
        decfsz  aux_count,F             ;loop count - 1
        goto    parse_call
        return
;------------------------------------------------------------
;       Assumes ssid_offset has been updated
parse_ssid                              ;saw space or 6
        movfw   ssid_offset             ;SSID loc to W
        movwf   RAMAddressLow           ;loc to pointer
        call    RAMReadByte             ;read byte at loc
        movwf   ssid_buffer             ;store in ssid buffer
        andlw   0x01                    ;test for S_D flag
        skpz
        bsf     SHIFTED_DATA            ;saw S_D flag
        rrf     ssid_buffer,F           ;right shift byte
        movlw   0x0f
        andwf   ssid_buffer,F           ;mask hi nibble
        skpnz
        return                          ;ssid = 0
        movlw   '-'                     ;add dash
        movwf   send_reg
        call    serial_out              ;send -
        movlw   0x0a                    ;10
        subwf   ssid_buffer,W           ;test magnitude
        skpc
        goto    send_the_ones           ;SSID < 10
        movwf   ssid_buffer             ;ssid buffer - 10
        movlw   0x31                    ;ascii 1 for tens
        movwf   send_reg                ;W to send_reg
        call    serial_out              ;send 10
send_the_ones
        movfw   ssid_buffer             ;ssid to W
        addlw   0x30                    ;ascii convert
        movwf   send_reg
        call    serial_out
        return                          ;done
;------------------------------------------------------------
;       Packet bit receive routine
;       Returns with 0 or 1 in Zero Flag
recv_one_bit                            ;delay 833 uS
        btfsc   INTCON,RBIF
        goto    bit_edge                ;saw tone toggle
        btfss   INTCON,T0IF
        goto    recv_one_bit            ;not timed out
        movlw   0x7e                    ;timed out
        movwf   TMR0                    ;reset timer
        bcf     INTCON,T0IF             ;1 byte long
        clrw                            ;Z flag set
        return
bit_edge
        movlw   0x3d                    ;load 1 1/2 bits
        movwf   TMR0
        movfw   PORTB                   ;reset port change
        bcf     INTCON,RBIF             ;clear RBIF
        bcf     INTCON,T0IF
        movfw   TMR0                    ;Z flag clear
        return
;------------------------------------------------------------
;               4800,8,N,1      serial out routine
;
serial_out
        movlw   0x08                    ;init shift counter
        movwf   serial_count
        bsf     SERIAL_DATA_OUT         ;start bit set
        call    onebit_delay
        comf    send_reg,F              ;invert data
        clrc
serial_out_loop
        rrf     send_reg,F              ;rotate bit 0 into carry
        skpnc                           ;test carry flag
        goto    hbit                    ;carry = 1
        bcf     SERIAL_DATA_OUT         ;bit is clear, output 0
        goto    time2_roll
hbit    bsf     SERIAL_DATA_OUT         ;bit is set, output 1
time2_roll
        call    onebit_delay
        decfsz  serial_count,F          ;shifted 8?
        goto    serial_out_loop         ;no
        bcf     SERIAL_DATA_OUT         ;yes, output stopbit
        call    onebit_delay
        return                          ;done
;------------------------------------------------------------
;                10 Mhz version
onebit_delay                            ;208 uS delay (4k8)
        movlw   0xad                    ;173 loops @ 1.2 uS
        movwf   delay_count
onebit_loop
        decfsz  delay_count,F
        goto    onebit_loop
        return
;------------------------------------------------------------
adjustable_delay                        ;W contains length
        movwf   aux_count
the_mS_loop
        call    onebit_delay
        decfsz  aux_count,F
        goto    the_mS_loop
        return
;------------------------------------------------------------
        END
